#include "mpdimpl.h"

bool ConnectAndRedirectInput(HANDLE hIn, char *pszHostPort, HANDLE hProcess, DWORD dwPid, int nRank)
{
    DWORD dwThreadID;
    RedirectSocketArg *pArg;
    int bfd;
    char pszHost[MAX_HOST_LENGTH];
    int nPort;
    char *pszPort;
    int nLength;
    char ch = 0;

    if ((pszHostPort == NULL) || (*pszHostPort == '\0'))
    {
	if (hIn != NULL)
	    CloseHandle(hIn);
	return true;
    }

    pszPort = strstr(pszHostPort, ":");
    if (pszPort == NULL)
    {
	if (hIn != NULL)
	    CloseHandle(hIn);
	return false;
    }
    nLength = pszPort - pszHostPort;
    strncpy(pszHost, pszHostPort, nLength);
    pszHost[nLength] = '\0';
    pszPort++;

    nPort = atoi(pszPort);

    if (beasy_create(&bfd, 0, INADDR_ANY) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectInput: beasy_create failed, error %d\n", WSAGetLastError());
	if (hIn != NULL)
	    CloseHandle(hIn);
	return false;
    }
    if (beasy_connect(bfd, pszHost, nPort) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectInput: beasy_connect(%s:%d) failed, error %d\n", pszHost, nPort, WSAGetLastError());
	beasy_closesocket(bfd);
	if (hIn != NULL)
	    CloseHandle(hIn);
	return false;
    }

    // Send header indicating stdin redirection
    if (beasy_send(bfd, &ch, sizeof(char)) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectInput: beasy_send(%d) failed, error %d\n", bget_fd(bfd), WSAGetLastError());
	beasy_closesocket(bfd);
	if (hIn != NULL)
	    CloseHandle(hIn);
	return false;
    }

    // Start thread to transfer data
    pArg = new RedirectSocketArg;
    pArg->hRead = NULL;
    pArg->bfdRead = bfd;
    pArg->hWrite = hIn;
    pArg->bfdWrite = BFD_INVALID_SOCKET;
    pArg->bReadisPipe = false;
    pArg->bWriteisPipe = true;
    pArg->hProcess = hProcess;
    pArg->dwPid = dwPid;
    pArg->nRank = nRank;
    pArg->cType = 0;
    HANDLE hThread;
    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RedirectSocketThread, pArg, 0, &dwThreadID);
    if (hThread == NULL)
    {
	err_printf("ConnectAndRedirectInput: CreateThread failed, error %d\n", GetLastError());
	CloseHandle(hIn);
	beasy_closesocket(bfd);
	return false;
    }
    else
	CloseHandle(hThread);
    return true;
}

bool ConnectAndRedirectOutput(HANDLE hOut, char *pszHostPort, HANDLE hProcess, DWORD dwPid, int nRank, char cType)
{
    DWORD dwThreadID;
    RedirectSocketArg *pArg;
    int bfd;
    char pszHost[MAX_HOST_LENGTH];
    int nPort;
    char *pszPort;
    int nLength;

    if ((pszHostPort == NULL) || (*pszHostPort == '\0'))
    {
	if (hOut != NULL)
	    CloseHandle(hOut);
	return true;
    }

    pszPort = strstr(pszHostPort, ":");
    if (pszPort == NULL)
    {
	if (hOut != NULL)
	    CloseHandle(hOut);
	return false;
    }
    nLength = pszPort - pszHostPort;
    strncpy(pszHost, pszHostPort, nLength);
    pszHost[nLength] = '\0';
    pszPort++;

    nPort = atoi(pszPort);
    
    if (beasy_create(&bfd, 0, INADDR_ANY) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectOutput: beasy_create failed, error %d\n", WSAGetLastError());
	if (hOut != NULL)
	    CloseHandle(hOut);
	return false;
    }
    if (beasy_connect(bfd, pszHost, nPort) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectOutput: beasy_connect(%s:%d) failed, error %d\n", pszHost, nPort, WSAGetLastError());
	beasy_closesocket(bfd);
	if (hOut != NULL)
	    CloseHandle(hOut);
	return false;
    }

    // Send header indicating stdout/err redirection
    if (beasy_send(bfd, &cType, sizeof(char)) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirectOutput: beasy_send(%d) failed, error %d\n", bget_fd(bfd), WSAGetLastError());
	beasy_closesocket(bfd);
	if (hOut != NULL)
	    CloseHandle(hOut);
	return false;
    }

    // Start thread to transfer data
    pArg = new RedirectSocketArg;
    pArg->hWrite = NULL;
    pArg->bfdWrite = bfd;
    pArg->hRead = hOut;
    pArg->bfdRead = BFD_INVALID_SOCKET;
    pArg->bReadisPipe = true;
    pArg->bWriteisPipe = false;
    pArg->hProcess = hProcess;
    pArg->dwPid = dwPid;
    pArg->nRank = nRank;
    pArg->cType = cType;
    HANDLE hThread;
    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RedirectSocketThread, pArg, 0, &dwThreadID);
    if (hThread == NULL)
    {
	err_printf("ConnectAndRedirectOutput: CreateThread failed, error %d\n", GetLastError());
	CloseHandle(hOut);
	beasy_closesocket(bfd);
	return false;
    }
    else
	CloseHandle(hThread);
    return true;
}

bool ConnectAndRedirect2Outputs(HANDLE hOut, HANDLE hErr, char *pszHostPort, HANDLE hProcess, DWORD dwPid, int nRank)
{
    DWORD dwThreadID;
    RedirectSocketArg *pArg;
    int bfd;
    char pszHost[MAX_HOST_LENGTH];
    int nPort;
    char *pszPort;
    int nLength;
    char cType = 1;

    if ((pszHostPort == NULL) || (*pszHostPort == '\0'))
    {
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return true;
    }

    pszPort = strstr(pszHostPort, ":");
    if (pszPort == NULL)
    {
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return false;
    }
    nLength = pszPort - pszHostPort;
    strncpy(pszHost, pszHostPort, nLength);
    pszHost[nLength] = '\0';
    pszPort++;

    nPort = atoi(pszPort);
    
    if (beasy_create(&bfd, 0, INADDR_ANY) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirect2Outputs: beasy_create failed, error %d\n", WSAGetLastError());
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return false;
    }
    if (beasy_connect(bfd, pszHost, nPort) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirect2Outputs: beasy_connect(%s:%d) failed, error %d\n", pszHost, nPort, WSAGetLastError());
	beasy_closesocket(bfd);
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return false;
    }

    // Send header indicating stdout/err redirection
    if (beasy_send(bfd, &cType, sizeof(char)) == SOCKET_ERROR)
    {
	err_printf("ConnectAndRedirect2Outputs: beasy_send(%d) failed, error %d\n", bget_fd(bfd), WSAGetLastError());
	beasy_closesocket(bfd);
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return false;
    }

    HANDLE hMutex = CreateMutex(NULL, FALSE, NULL);
    HANDLE hThread;

    // Start thread to transfer data from hOut
    pArg = new RedirectSocketArg;
    pArg->hWrite = NULL;
    pArg->bfdWrite = bfd;
    pArg->hRead = hOut;
    pArg->bfdRead = BFD_INVALID_SOCKET;
    pArg->bReadisPipe = true;
    pArg->bWriteisPipe = false;
    pArg->hProcess = hProcess;
    pArg->dwPid = dwPid;
    pArg->hMutex = hMutex;
    pArg->bFreeMutex = false;
    pArg->nRank = nRank;
    pArg->cType = 1;
    pArg->hOtherThread = NULL;
    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RedirectLockedSocketThread, pArg, 0, &dwThreadID);
    if (hThread == NULL)
    {
	err_printf("ConnectAndRedirect2Outputs: CreateThread failed, error %d\n", GetLastError());
	if (hOut != NULL)
	    CloseHandle(hOut);
	beasy_closesocket(bfd);
	if (hErr != NULL)
	    CloseHandle(hErr);
	return false;
    }

    // Start thread to transfer data from hErr
    pArg = new RedirectSocketArg;
    pArg->nRank = nRank;
    pArg->hWrite = NULL;
    pArg->bfdWrite = bfd;
    pArg->hRead = hErr;
    pArg->bfdRead = BFD_INVALID_SOCKET;
    pArg->bReadisPipe = true;
    pArg->bWriteisPipe = false;
    pArg->hProcess = NULL;
    pArg->dwPid = -1;
    pArg->hMutex = hMutex;
    pArg->bFreeMutex = true;
    pArg->nRank = nRank;
    pArg->cType = 2;
    pArg->hOtherThread = hThread;
    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RedirectLockedSocketThread, pArg, 0, &dwThreadID);
    if (hThread == NULL)
    {
	err_printf("ConnectAndRedirect2Outputs: CreateThread failed, error %d\n", GetLastError());
	CloseHandle(pArg->hOtherThread);
	if (hOut != NULL)
	    CloseHandle(hOut);
	if (hErr != NULL)
	    CloseHandle(hErr);
	beasy_closesocket(bfd);
	return false;
    }
    else
	CloseHandle(hThread);
    return true;
}
